from __future__ import (absolute_import, division,print_function)
import maya.cmds as cmds
import maya.api.OpenMaya as om

import os
import json
from xml.etree import ElementTree
from . import rigSettings

##----------------------------------------------------------------------------------------
##editStrings
##----------------------------------------------------------------------------------------
def capitalizeString(sourceString):
    distString = sourceString[0].upper() + sourceString[1:]
    return distString

def addPrefix(prefix,sourceString):    
    if sourceString == "":
        return prefix    
    if prefix != "":
        return prefix + capitalizeString(sourceString)
    else:
        return sourceString

def getSideString(side):
    sideString = ""
    
    if side in ["L","left","Left"]:
        sideString = "_L"

    if side in ["R","right","Right"]:
        sideString = "_R"
    
    return sideString

##----------------------------------------------------------------------------------------
##editFiles
##----------------------------------------------------------------------------------------
def readJSON(filePath,defaultDict = None):
    importDict = {}

    if os.path.exists(filePath) == True and os.path.isfile(filePath):
        with open(filePath) as f:
            importDict = json.load(f)	

    if defaultDict == None:
        return importDict

    else:
        for key in list(defaultDict.keys()):
            if key not in list(importDict.keys()):
                importDict[key] = defaultDict[key]

    return importDict

def writeJSON(filePath,valueDict,indent = 4):	
    with open(filePath, 'w') as f:
        json.dump(valueDict, f, indent = indent)

##----------------------------------------------------------------------------------------
##editAttrs
##----------------------------------------------------------------------------------------
def getFlattenAttr(target,attrs = []):
    targetAttrs = []
    for attr in attrs:
        if cmds.attributeQuery(attr,node = target,exists =True) ==False:
            if cmds.objExists(target + "." + attr):
                targetAttrs.append(attr)
            continue

        childrenAttrs = cmds.attributeQuery(attr,node = target,listChildren =True)        
        
        if childrenAttrs == None or len(childrenAttrs) == 0:
            targetAttrs.append(attr)
        else:
            targetAttrs.extend(childrenAttrs)

    return targetAttrs

def setAttrState(targets,attrs,lock,hide,keyable):
    for target in targets:
        targetAttrs = getFlattenAttr(target,attrs)
        for attr in targetAttrs:
            if cmds.objExists(target + "." + attr) == False:
                continue
            #lock
            cmds.setAttr(target + '.'+ attr, lock = lock)

            if hide:
                cmds.setAttr(target + '.'+ attr,  channelBox = False, keyable = False)
            else :
                cmds.setAttr(target + '.'+ attr, channelBox = True, keyable = True)
                cmds.setAttr(target + '.'+ attr, keyable = keyable)

def connectAttrZip(source,target,sourceAttrs,targetAttrs,f =True):
    sourceAttrs = getFlattenAttr(source,sourceAttrs)
    targetAttrs = getFlattenAttr(target,targetAttrs)
    
    for sourceAttr,targetAttr in zip(sourceAttrs,targetAttrs):           
        cmds.connectAttr(source + "." + sourceAttr,target + "." + targetAttr, f =f)

def listAttrs(node,keyable,unkeyable,longName = False):
    attrs = []

    unKeyableAttrs = cmds.listAttr(node,channelBox =True) or []
    keyableAttrs = cmds.listAttr(node,keyable =True) or []
    
    if keyable:

        for attr in keyableAttrs:
            attrName = cmds.attributeQuery(attr, node=node, longName=longName,  shortName = not longName)
            attrs.append(attrName)

    if unkeyable:
        for attr in unKeyableAttrs:
            attrName = cmds.attributeQuery(attr, node=node, longName=longName,  shortName = not longName)
            attrs.append(attrName)

    return attrs

def getTransformLimit(node):
    limitDict = {}

    for attr in ["Trans","Rot","Scale"]:
        for axis in "XYZ":
            if cmds.getAttr(node + ".min"+attr+axis+"LimitEnable"):
                limitDict["min"+attr+axis+"LimitEnable"] = True
                limitDict["min"+attr+axis+"Limit"] = cmds.getAttr(node + ".min"+attr+axis+"Limit")
            
            if cmds.getAttr(node + ".max"+attr+axis+"LimitEnable"):
                limitDict["max"+attr+axis+"LimitEnable"] = True
                limitDict["max"+attr+axis+"Limit"] = cmds.getAttr(node + ".max"+attr+axis+"Limit")

    return limitDict
##----------------------------------------------------------------------------------------
##listTransformNodes
##----------------------------------------------------------------------------------------
def getCurHierarchy(target,topNode = None):
    hierarchyList = []
    hierarchyList.append(target)

    if target == topNode:
        return hierarchyList

    reachTop = False    
    while reachTop == False:
        parent = cmds.listRelatives(hierarchyList[-1], p =True)
        
        if parent == None:
            reachTop = True
            continue

        elif parent[0] == topNode:
            reachTop = True

        hierarchyList.append(parent[0])
    hierarchyList.reverse()

    return hierarchyList

def listSetMembers(setName,nodeTypes,allHierarchy):
    if cmds.objExists(setName) ==False:
        return []

    allMembers = cmds.sets(setName,q=True)
    memberSets = cmds.ls(allMembers, type = "objectSet") or []
    memberNodes = []

    for nodeType in nodeTypes:
        nodes = cmds.ls(allMembers, type = nodeType) or []
        memberNodes = list(set(memberNodes) | set(nodes))

    if allHierarchy:
        ##setMemberが objectSetの場合、掘り下げる
        for memberSet in memberSets:
            memberNodes.extend(listSetMembers(memberSet,nodeTypes,allHierarchy))
    
    return memberNodes

def listIKJoints():    
    ikHandles = cmds.ls(type = "ikHandle") or []

    IKJoints = []
    for ikHandle in ikHandles:
        startJoint = cmds.listConnections(ikHandle + ".startJoint",s=True, d=False) or []
        endEffector = cmds.listConnections(ikHandle + ".endEffector",s=True, d=False) or []
        endJoint = cmds.listConnections(endEffector[0] + ".offsetParentMatrix",s=True, d=True) or []
        jointChain = getCurHierarchy(endJoint,startJoint)
        IKJoints = list(set(IKJoints) | set(jointChain))

    return IKJoints

def listNodes(topNode,ctrlSets):
    ##階層下全てのtransformノードを抽出
    allTransformNodes = cmds.listRelatives(topNode,ad =True, type ="transform") or []
    allTransformNodes.append(topNode)

    ##コンストレインノードを抽出
    constraintNodes = cmds.listRelatives(topNode,ad =True, type ="constraint") or []
    otherNodes = list(set(allTransformNodes) - set(constraintNodes))

    ##IKhandleで制御されているjointはlockしてしまうと正しく動かなくなるので、別途抽出    
    IKJoints = listIKJoints()
    otherNodes = list(set(otherNodes) - set(IKJoints))

    ##ctrlsetsに登録されているノードを抽出
    ctrlNodes = listSetMembers(ctrlSets,["transform"],True)
    otherNodes = list(set(otherNodes) - set(ctrlNodes))
    
    nodeDict = {
                "ctrl":ctrlNodes,
                "IKJoints":IKJoints,
                "constraint":constraintNodes,
                "other":otherNodes
    }

    return nodeDict


##----------------------------------------------------------------------------------------
##edit Transform
##----------------------------------------------------------------------------------------
def getPosition(target):
    nodeType = cmds.nodeType(target)
    
    if nodeType == "joint" or nodeType == "transform" or nodeType == "ikHandle" or nodeType == "place3dTexture":
        # return cmds.xform(target , q =True , ws = True, t =True)
        return cmds.xform(target , q =True , ws = True, rotatePivot =True)

    elif nodeType == "mesh" or nodeType == "nurbsCurve" or nodeType == "nurbsSurface" or nodeType == "lattice":
        return cmds.xform(target , q =True , ws = True, t =True)

def getRotation(target):
    return cmds.xform(target , q =True , ws = True, ro =True)

def snapPosition(source,target):
    destPosition = getPosition(source)
    cmds.xform(target, ws =True, t =destPosition)

    return

def snapRotation(source,target):
    rotation = getRotation(source)
    cmds.xform(target, ws =True, ro = rotation)

    if cmds.nodeType(target) == "joint":
        cmds.makeIdentity(target,apply = True,r =1)
    return

def getDagNode(target):    
    try:
        sellist = om.MGlobal.getSelectionListByName(target)
        return sellist.getDagPath(0)
    except:
        return None
    
def getShapeFn(dagPath):
    FnShape = None

    if dagPath.hasFn(om.MFn.kMesh):
        FnShape = om.MFnMesh(dagPath)        

    elif dagPath.hasFn(om.MFn.kNurbsSurface):
        FnShape = om.MFnNurbsSurface(dagPath)

    elif dagPath.hasFn(om.MFn.kNurbsCurve):
        FnShape = om.MFnNurbsCurve(dagPath)

    return FnShape
##----------------------------------------------------------------------------------------
##for controller
##----------------------------------------------------------------------------------------
def getCVPositions(target,space = om.MSpace.kWorld):
    dagPath = getDagNode(target)
    FnShape = getShapeFn(dagPath)
    
    cvPositions = []
    if FnShape == None:
        return cvPositions

    for position in FnShape.cvPositions(space):
        cvPositions.append(list(position)[:3])

    return cvPositions

def readCurveShape(target,space = "world"):
    if space == "world":
        space = om.MSpace.kWorld
    elif space == "object":
        space = om.MSpace.kObject

    cvPositions = getCVPositions(target,space = space)
    dagPath = getDagNode(target)
    
    form = None
    knots = None
    degree = None
    FnShape = getShapeFn(dagPath)
    if FnShape != None:
        form = FnShape.form
        knots = list(FnShape.knots())
        degree = FnShape.degree 
        
    return cvPositions,degree,form,knots

def getShapeColorGRB(target):
    shapes = cmds.listRelatives(target,s =True)
    rgb = [0.0,0.0,0.0]

    if shapes == None:
        return rgb

    if cmds.getAttr(shapes[0] + '.useObjectColor') == 2:
        rgb = cmds.getAttr(shapes[0] + '.wireColorRGB')[0]

    return rgb

def setShapeColorGRB(target,RGB):
    shapes = cmds.listRelatives(target,s =True) or []

    for shape in shapes:		
        cmds.setAttr(shape + '.useObjectColor', 2)
        cmds.setAttr(shape + '.wireColorRGB', RGB[0],RGB[1],RGB[2], type = 'double3')

def addWireSettingKeep(target):
    shapes = cmds.listRelatives(target,s =True) or []
    if len(shapes) == 0:
        return

    wireColor = getShapeColorGRB(target)

    if cmds.objExists(target + ".wireColorKeep") == False:
        cmds.addAttr(target, ln = "wireColorKeep",at = "compound",numberOfChildren = 3)
        cmds.addAttr(target, ln = "wireColorKeepR",at = "float", p = "wireColorKeep",dv = wireColor[0], min = 0.0,max = 1.0)
        cmds.addAttr(target, ln = "wireColorKeepG",at = "float", p = "wireColorKeep",dv = wireColor[1], min = 0.0,max = 1.0)
        cmds.addAttr(target, ln = "wireColorKeepB",at = "float", p = "wireColorKeep",dv = wireColor[2], min = 0.0,max = 1.0)

    for shape in shapes:		    
        cmds.setAttr(target + ".wireColorKeepR",l =False)
        cmds.setAttr(target + ".wireColorKeepG",l =False)
        cmds.setAttr(target + ".wireColorKeepB",l =False)
        
        cmds.connectAttr(target + ".wireColorKeepR",shape + ".wireColorR")
        cmds.connectAttr(target + ".wireColorKeepG",shape + ".wireColorG")
        cmds.connectAttr(target + ".wireColorKeepB",shape + ".wireColorB")

def setWireFrameColor(targets,color,tone = 1.0):
    RGB = rigSettings.COLOR_DICT[color]    
    RGB = [RGB[0]*tone,RGB[1]*tone,RGB[2]*tone]

    for target in targets:
        setShapeColorGRB(target,RGB)

def generateKnot(degree,cvNum):
    knot = om.MDoubleArray()
    
    if degree == 1:
        for i in range(0,cvNum):
            knot.append(i)
    
    elif degree == 2:		
        knotCount = 0
                
        for i in range(0,cvNum+1):
            knot.append(knotCount)
        
            if i > 0 and i < (cvNum-1):
                knotCount = knotCount +1
    
    elif degree == 3:
        knotCount = 0
                
        for i in range(0,cvNum+2):
            knot.append(knotCount)
    
            if i > 1 and i < (cvNum-1):
                knotCount = knotCount +1
    
    return list(knot)

def createCurveShape(cvPositions,degree,form,knots = None,parent = None,editPoint = False):
    if knots == None:
        knots = generateKnot(degree,len(cvPositions))

    periodic = False
    if form == 3:
        periodic = True
    
    if editPoint ==False:
        curve = cmds.curve(point = cvPositions, knot = knots, degree = degree,periodic = periodic)
    else:
        curve = cmds.curve(ep = cvPositions, knot = knots, degree = degree,periodic = periodic)
    
    if parent != None:        
        shape = cmds.listRelatives(curve, type = "shape")[0]

        if cmds.objExists(parent) == False:
            cmds.createNode("transform",name = parent)

        cmds.parent(shape,parent,s=True,r=True)
        cmds.rename(shape,parent + "Shape")
        cmds.delete(curve)
        
        curve = parent

    return curve

def createCtrlNode(nodeName,shapeName,nodeType,size = 1.0,position = None,rotation =None):
    transformNode = cmds.createNode(nodeType,name = nodeName)
    
    transformNode = createCurveShape(
                                        rigSettings.CURVE_SHAPE_DICT[shapeName]["cvs"],
                                        rigSettings.CURVE_SHAPE_DICT[shapeName]["degree"],
                                        rigSettings.CURVE_SHAPE_DICT[shapeName]["form"],
                                        rigSettings.CURVE_SHAPE_DICT[shapeName]["knots"],parent = transformNode)

    transformNode = cmds.rename(transformNode,nodeName,ignoreShape =False)
    cmds.xform(transformNode, scale = [size,size,size], ws = True)
    cmds.makeIdentity(transformNode, a =True, r=True, s =True)

    if position != None:
        snapPosition(position,transformNode,False)

    if rotation != None:
        snapRotation(rotation,transformNode,False)

    if nodeType == "joint":
        cmds.setAttr(transformNode + ".drawStyle",2)
        setAttrState([transformNode],["radius"],lock=False,hide=True,keyable=False)

    for shape in cmds.listRelatives(transformNode,type = "shape"):
        cmds.setAttr(shape + ".ihi",False)

    return transformNode

def addToSet(setName,nodes):
    if cmds.objExists(setName) == False:
        cmds.sets(name = setName, empty=True)

    cmds.sets(nodes,e=True,add = setName)

    
##----------------------------------------------------------------------------------------
##skincluster
##----------------------------------------------------------------------------------------
def getSkinCluster(node):
    history = cmds.listHistory(node,pruneDagObjects =True)    
    skinNode = cmds.ls(history,type = "skinCluster") or None
    
    if skinNode != None:
        return skinNode[0]

    return None

def removeUnuseInfs(skinNode):
    allInf =  cmds.skinCluster(skinNode,q =True, influence =True)
    weightedInf= cmds.skinCluster(skinNode,q =True, weightedInfluence =True)
    removeInf = list(set(allInf) - set(weightedInf))

    for inf in removeInf:
        cmds.skinCluster(skinNode,e =True, removeInfluence = inf)

def exportSkinWeight(dirPath,node,ext = "xml",cleanup = True,makeDir = True):
    if cmds.nodeType(node) != "transform":
        return
    
    ##skinClusterノードを取得
    skinNode = getSkinCluster(node)
    if skinNode == None:
        return

    ##wieghtが振られていないinfluenceを除去
    if cleanup:
        removeUnuseInfs(skinNode)

    ##skinClsuter名がユニークになるように編集
    if skinNode != node + rigSettings.SUFFIX_DICT["skinCluster"]:
        skinNode = cmds.rename(skinNode, node + rigSettings.SUFFIX_DICT["skinCluster"])

    if fileName == None:
        fileName = node + "_skinWeight." + ext
    else:
        fileName = fileName + "." + ext

    if os.path.exists(dirPath) ==False:
        os.makedirs(dirPath)
        
    cmds.deformerWeights(fileName, path = dirPath, deformer = skinNode, format = ext.upper(),ex = True)

    return fileName

def createSkinCluster(target,joints,nodeName = None):
    if nodeName == None:
        nodeName = target + rigSettings.SUFFIX_DICT["skinCluster"]

    if cmds.objExists(nodeName) ==False:
        cmds.skinCluster(joints,target,normalizeWeights = 1,toSelectedBones =True,name = nodeName)[0]

    return nodeName

def getInfluence(skinNode):
    infs = cmds.skinCluster(skinNode, q =True, inf =True)    
    return infs

def addInfluence(target,joints):
    skinNode = getSkinCluster(target)

    if skinNode == None:
        skinNode = createSkinCluster(target,joints)

    curInfs = getInfluence(skinNode)
    toAddInfs = list(set(joints) - set(curInfs))
    
    for inf in toAddInfs:
        cmds.skinCluster(skinNode, e =True, ai = inf, weight = 0)

    return skinNode

def importSkinWeight(path,fileName):
    allInfsDict = {}
    
    ##XML　からバインド情報を取得
    tree = ElementTree.parse(path + fileName)
    root = tree.getroot()    
    for e in root.findall('weights'):
        inf = e.get("source")
        shape = e.get("shape")
        node = e.get("deformer")

        if node in allInfsDict.keys():
            allInfsDict[node]["infs"].append(inf)
        else:
            allInfsDict[node] = {
                                    "infs":[inf],
                                    "shape":shape
            }

    for skinNode in allInfsDict.keys():
        allInfs = allInfsDict[node]["infs"]
        shape = allInfsDict[node]["shape"]

        ##skinClusterが存在しなければ作成し、インフルエンスを追加する
        skinNode = addInfluence(shape,allInfs)

        if len(allInfs) != 1:
            cmds.deformerWeights(fileName, path = path, deformer = skinNode, method = "index",im = True)
            cmds.skinCluster(skinNode,e =True,forceNormalizeWeights = True)